import uuid
from fastapi import APIRouter, Header, Request
from starlette.exceptions import HTTPException as StarletteHTTPException
from pydantic import BaseModel
from typing import Optional, Union
import secrets

router = APIRouter()

def abort(code: int) -> StarletteHTTPException:
    raise StarletteHTTPException(status_code=code)

def secure_strcmp(val1, val2):
    """
    From Django:

    Return True if the two strings are equal, False otherwise. This is a secure function
    """
    return secrets.compare_digest(val1, val2)


class VoteContext(BaseModel):
    """
        Represents a fateslist vote context. fateslist.py will make this a Vote class
    """
    user: str
    votes: int
    test: Optional[bool] = False

class Event(BaseModel):
    """Represents a event on fateslist"""
    e: int
    eid: uuid.UUID
    t: int
    ts: float
    user: str

class VoteModel(BaseModel):
    """The vote information itself"""
    ctx: VoteContext
    id: str
    m: Event

class Vote():
    """
        Represents a vote on IBL

        :param bot_id: The Bot ID of the vote

        :param user_id: The ID of the user who voted for your bot. In test mode, this will be 0

        :param username: The username who voted for your bot

        :param count: The amount of votes your bot now has

        :param test: Whether this is a test webhook or not

        :param timestamp: The timestamp (epoch) when the vote happened
    """
    def __init__(self, bot_id: int, user_id: int, test: bool, timestamp: int, count: int, username: str):
        self.bot_id = bot_id
        self.user_id = user_id
        self.test = test
        self.timestamp = timestamp

        if count.isdigit():
            self.count = int(count)
        else:
            self.count = 0

@router.post("/_dbg")
async def debug_webhook(request: Request):
    print((await request.body()), secret)

@router.post("/")
async def iblpy_webhook(vote: VoteModel, Authorization: str = Header("INVALID_SECRET")):
    if secret is None or secure_strcmp(secret, Authorization):
        pass
    else:
        return abort(401)

    return await wh_func(vote, secret)


import aiohttp
from .http import BaseHTTP
from .classes import Stats, Vanity

class SystemClient(BaseHTTP):
    def __init__(self):
        self.id = None # To ensure anything using id doesnt just die
        self.logged_in = False 

    async def blstats(self):
        """
        Returns bot list stats
        """
        api_res = await self.request(
            method="GET",
            endpoint="/blstats",
        )

        if not api_res.success:
            return api_res
        return Stats(api_res.json)
    
    async def get_vanity(self, code: str):
        """
        Returns the vanity url for a given code
        """
        api_res = await self.request(
            method="GET",
            endpoint=f"/code/{code}",
        )

        if not api_res.success:
            return api_res
        return Vanity(api_res.json)


import fateslist.config as cfg
from fateslist.classes import APIResponse, APIRatelimit, Bot, User, UserVotes
import aiohttp
from typing import Optional
from loguru import logger
try:
    import orjson as json_lib
except:
    logger.warning("Orjson could not be imported... falling back to json. Performance will be decreased")
    import json as json_lib

class BaseHTTP():
    user_agent = f"IBLPy/{cfg.version}"       
    
    def __init__(self, id: int):
        self.id = id
        self.logged_in = False 

    def login(self, api_token):
        """Logs in to the IBL API"""
        self.api_token = api_token
        self.logged_in = True
    
    async def request(
        self, 
        method: str, 
        endpoint: str,
        json: Optional[dict] = None, 
        headers: Optional[dict] = None, 
        auth: bool = False
    ):
        """Makes a API request"""
        headers = {} if not headers else headers
        
        if auth:
            if not self.logged_in:
                raise ValueError("Not logged in to IBL yet")
                
            headers["authorization"] = self.api_token
        
        headers["User-Agent"] = self.user_agent
        
        async with aiohttp.ClientSession() as sess:
            f = getattr(sess, method.lower())
            async with f(f"{cfg.api}{endpoint}", headers = headers, json = json) as res:
                if res.status == 429:
                    raise APIRatelimit()
           
                # Convert it to json
                i = 0
                json = await res.text()
                while isinstance(json, (bytes, str)) and i < 100:
                    json = json_lib.loads(json)
                    i+=1
            
                logger.debug(str(json))
            
                return APIResponse(res = res, json = json)
     
    
class BotHTTP(BaseHTTP):    
    async def set_stats(
        self,
        guild_count: int, 
        shard_count: Optional[int] = None
    ):        
        return await self.request(
            method="POST",
            endpoint=f"/bots/{self.id}/stats",
            json={
                "guild_count": guild_count,
                "shard_count": shard_count if shard_count else "0" # IBL API workaroud
            },
            auth=True
        )
   
    async def get_bot(self):
        api_res = await self.request(
            method="GET",
            endpoint=f"/bots/{self.id}",
        )
        
        if not api_res.success:
            return api_res
        
        return Bot(api_res.json)

    async def get_user_votes(self, user_id: int):
        api_res = await self.request(
            method="GET",
            endpoint=f"/users/{user_id}/bots/{self.id}/votes",
            auth=True
        )

        if not api_res.success:
            return api_res
        return UserVotes(api_res.json)

    
class UserHTTP(BaseHTTP):         
    async def get_user(self, debug: bool = False):
        api_res = await self.request(
            method="GET",
            endpoint=f"/users/{self.id}", 
        )

        if not api_res.success:
            return api_res
    
        return User(api_res.json)




import fateslist.config as cfg
from fateslist import api_modes, aiohttp
from typing import Union, Optional
from discord import Embed
from .utils import extract_time
import datetime

class InvalidMode(Exception):
    """Raised when you don't have the required mode (package) to perform the action such as trying to do an asynchronous API request without having aiohttp_requests installed or trying to do a webhook without fastapi+uvicorn"""
    def __init__(self, mode):
        if mode == "async":
            super().__init__("In order to use fateslist asynchronous API requests, you must have aiohttp, requests and aiohttp_requests installed")
        elif mode == "fastapi":
            super().__init__("In order to use fateslist webhooks, you must have fastapi and uvicorn installed")

class APIRatelimit(Exception):
    """Raised when you are being ratelimited by IBL. The ratelimit for posting stats is 3 requests per 5 minutes and is unknown/variable for getting stats from the API"""
    def __init__(self):
        super().__init__("You are being ratelimited by the Infinity Bots (IBL) API. For future reference, the ratelimit for posting stats is 3 requests per 5 minutes and is unknown/variable for getting stats from the API!")

class APIResponse():
    """
        APIResponse represents an API response in the fateslist library
        
        :param res: This is the raw response from the API. 
            This will be a aiohttp ClientResponse

        :param done: Whether the API response has succeeded or not

        :param success: If the API response has succeeded, this will be false. Similar to done but based on status code
        
        :param reason: The error message reported by the Fates List API

        :param message: Any messages returned by the API in the message field. Can be None if there are no messages

        :param json: The JSON object sent by the API

        :param status: The status code of the HTTP response received from the API
    """
    def __init__(self, *, res: aiohttp.ClientResponse, json: dict):
        self.res = res
        self.done = json.get("done")
        self.success = res.status < 400
        self.reason = json.get("reason")
        self.json = json
        self.status = res.status

class BaseObject():
    def __init__(self, json):
        self.__dict__.update(**json)
    
    def dict(self) -> dict:
        """Returns the class as a dict using the dict dunder property of the class"""
        return self.__dict__

class BaseUser(BaseObject):
    """
        This is a base user on fateslist from which all bots and users extend from
    """
    def __str__(self) -> str:
        """Returns the name of the bot or user"""
        try:
            return self.user.username
        except AttributeError:
            return str(self.dict())

    def __int__(self) -> int:
        """Returns the bot or user ID"""
        return self.user.id
    
class Bot(BaseUser):
    """
        Bot is internally a part of the classes module (which provides all of fateslist's base classes and functions). 
        It represents a bot on Fates List. The exact parameters of an bot may change and fateslist is designed to handle such changes automatically. 

        Please see https://api.fateslist.xyz/api/docs/redoc#operation/fetch_bot for a full list of parameters.

        You should access parameters using object notation
    """
    ...

class User(BaseUser):
    """
        User is internally a part of the classes module (which provides all of fateslist's base classes and functions). 
        It represents a user on Fates List. The exact parameters of an user may change and fateslist is designed to handle such changes automatically. 

        Please see https://api.fateslist.xyz/api/docs/redoc#operation/fetch_user for a full list of parameters.

        You should access parameters using object notation
    """
    ...

class UserVotes(BaseObject):
    """
        User is internally a part of the classes module (which provides all of fateslist's base classes and functions). 
        It represents user votes on Fates List. The exact parameters of an user votes may change and fateslist is designed to handle such changes automatically. 

        Please see https://api.fateslist.xyz/api/docs/redoc#operation/get_user_votes for a full list of parameters.

        You should access parameters using object notation
    """
    ...

class Stats(BaseObject):
    """
        Stats is internally a part of the classes module (which provides all of fateslist's base classes and functions). 
        It represents stats on Fates List. The exact parameters of stats may change and fateslist is designed to handle such changes automatically. 

        Please see https://api.fateslist.xyz/api/docs/redoc#operation/blstats for a full list of parameters.

        You should access parameters using object notation
    """
    def embed(self):
        """
        Returns a embed of fates list stats
        """
        embed = Embed(title="Bot List Stats", description="Fates List Stats")
        uptime_tuple = extract_time(datetime.timedelta(seconds=self.uptime))
        # ttvr = Time Till Votes Reset
        ttvr_tuple = extract_time(
            (datetime.datetime.now().replace(day=1, second=0, minute=0, hour=0) +
            datetime.timedelta(days=32)).replace(day=1) - datetime.datetime.now())
        uptime = "{} days, {} hours, {} minutes, {} seconds".format(*uptime_tuple)
        ttvr = "{} days, {} hours, {} minutes, {} seconds".format(*ttvr_tuple)
        embed.add_field(name="Uptime", value=uptime)
        embed.add_field(name="Time Till Votes Reset", value=ttvr)
        embed.add_field(name="Worker PID", value=str(self.pid))
        embed.add_field(name="Worker Number",
                        value=self.workers.index(self.pid) + 1)
        embed.add_field(
            name="Workers",
            value=f"{', '.join([str(w) for w in self.workers])} ({len(self.workers)} workers)",
        )
        embed.add_field(name="UP?", value=str(self.up))
        embed.add_field(name="Server Uptime", value=str(self.server_uptime))
        embed.add_field(name="Bot Count", value=str(self.bot_count))
        embed.add_field(name="Bot Count (Total)",
                        value=str(self.bot_count_total))
        return embed

class Vanity(BaseObject):
    """
        Vanoty is internally a part of the classes module (which provides all of fateslist's base classes and functions). 
        It represents vanities (slugs) on Fates List. The exact parameters of stats may change and fateslist is designed to handle such changes automatically. 

        Please see https://api.fateslist.xyz/api/docs/redoc#operation/get_vanity for a full list of parameters.

        You should access parameters using object notation
    """
    ...


import datetime
import traceback

def extract_time(td: datetime.timedelta) -> tuple:
    """
    Extracts time from a timedelta object.
    
    usage: (_d, _h, _m, _s, _mils, _mics) = tdTuple(td)
    """
    def _t(t, n):
        if t < n:
            return (t, 0)
        v = t // n
        return (t - (v * n), v)

    (s, h) = _t(td.seconds, 3600)
    (s, m) = _t(s, 60)
    return (td.days, h, m, s)

def etrace(ex):
    """
    Get the traceback of an exception

    Used internally by fateslist.py and our own tooling
    """
    try:
        return "".join(traceback.format_exception(ex)) # COMPAT: Python 3.10 only
    except:
        return str(ex)



api_modes = []
try:
    import aiohttp
    api_modes.append("async")
except Exception:
    raise ImportError("You must have aiohttp installed to import this library!")
    
try:
    import uvicorn
    import fastapi
    api_modes.append("webhook")
except:
    uvicorn = None
    fastapi = None

import discord

from fateslist.main import *
from fateslist.config import *
from fateslist.ws import *
from fateslist.utils import *
from fateslist.system import *

__version__ = version
